#pragma once

#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 200112L
#endif

#include<stddef.h>
#include<stdint.h>

#ifdef _WIN32
#include<winsock2.h>
#include<ws2tcpip.h>
#else
#include<netdb.h>
#endif

/* Standalone hole punching libary (IPv4 & IPv6 compatible) */

#define STOON_CONN_INFO_SIZE (4 + 2 + 16 + 2)

#define POONCH_MAGIC 0x504F4348
#define POONCH_STAGE_KNOCKING 0
#define POONCH_STAGE_ACK 1
struct Poonch {
	uint32_t magic;
	uint32_t stage;
};

struct Stoon {
	struct sockaddr_in stu4;
	struct sockaddr_in my4;
	int fd4;
	
	struct sockaddr_in6 stu6;
	struct sockaddr_in6 my6;
	int fd6;
	
	uint8_t poonchstage;
};

// stunhost: Self-explanatory
// stunport: Self-explanatory
// myport: Local endpoint port
struct Stoon stoon_init(const char *stunhost, uint16_t stunport, uint16_t myport);

int stoon_req(struct Stoon*);
int stoon_listen(struct Stoon*);

// This is necessary in order to keep the NAT table entry active
// Some NATs wait as little as 25 seconds, so an interval *less* than
// 25s is recommended
void stoon_keepalive(struct Stoon*);

void stoon_kill(struct Stoon*);

int stoon_v4_available(struct Stoon*);
int stoon_v6_available(struct Stoon*);

void stoon_serialize(struct Stoon*, uint8_t ret[static STOON_CONN_INFO_SIZE]);

#define STOON_POONCH_INCOMPATIBLE_NETFAMS (-1)
#define STOON_POONCH_NO_KNOCK 0
#define STOON_POONCH_ACKED 1
int stoon_poonch(struct Stoon*, const uint8_t peerdata[static STOON_CONN_INFO_SIZE]);
